from __future__ import annotations

from dictionizr import dictionize, undictionize

from .utils import eq


class ComplexObject:
    def __init__(self,
        required: str,
        *args,
        optional: str = '',
        **kwargs
    ) -> None:
        self.required = required
        self.args = args
        self.optional = optional
        self.kwargs = kwargs
    def __eq__(self, o: ComplexObject) -> bool:
        return eq(self, o)


def test__dictionize__complex_object():
    data = ComplexObject(
        'required',
        'arg1',
        'arg2',
        optional='optional',
        hello='world',
        hola='mundo',
    )
    actual = dictionize(data)
    expected = {
        'required': 'required',
        'optional': 'optional',
        'args': [
            'arg1',
            'arg2',
        ],
        'kwargs': {
            'hello': 'world',
            'hola': 'mundo',
        }
    }
    assert expected == actual


def test__undictionize__complex_object():
    data = {
        'required': 'required',
        'optional': 'optional',
        'args': [
            'arg1',
            'arg2',
        ],
        'kwargs': {
            'hello': 'world',
            'hola': 'mundo',
        }
    }
    actual = undictionize(data, ComplexObject)
    expected = ComplexObject(
        'required',
        'arg1',
        'arg2',
        optional='optional',
        hello='world',
        hola='mundo',
    )
    assert expected == actual



from __future__ import annotations
import inspect
from typing import Optional, Type
from types import SimpleNamespace

def dictionize(data) -> dict:
    if data is None:
        return {}

    try:
        output = vars(data)
    except TypeError:
        output = {}
        for slot in data.__slots__:
            try:
                output[slot] = getattr(data, slot)
            except AttributeError:
                pass

    for key, value in output.items():
        if hasattr(value, '__dict__'):
            output[key] = dictionize(value)
        elif isinstance(value, tuple):
            output[key] = list(value)

    output = {
        key: value
        for key, value
        in output.items()
        if value is not None
    }
    return output


def undictionize(data: dict, class_: Optional[Type] = None):
    data = data.copy()
    if data is None:
        return None

    if class_ is None:
        obj = SimpleNamespace()
    else:
        obj = class_.__new__(class_)

    for key, value in data.items():
        if isinstance(value, dict):
            value = undictionize(value)
        setattr(obj, key, value)

    if class_ is not None and hasattr(class_.__init__, '__code__'):
        init_signature = inspect.signature(class_.__init__)
        init_parameters = init_signature.parameters
        positional_args = []
        keyword_args = {}
        for param in init_parameters.values():
            if param.name == 'self':
                continue

            # See: https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
            if param.kind == param.POSITIONAL_OR_KEYWORD or param.kind == param.POSITIONAL_ONLY:
                value = data.pop(param.name, param.default)
                positional_args.append(value)
            elif param.kind == param.KEYWORD_ONLY:
                value = data.pop(param.name, param.default)
                keyword_args[param.name] = value
            elif param.kind == param.VAR_POSITIONAL: # args
                if 'args' in data and isinstance(data['args'], list):
                    positional_args.extend(data['args'])
            elif param.kind == param.VAR_KEYWORD: # kwargs
                if 'kwargs' in data and isinstance(data['kwargs'], dict):
                    keyword_args.update(**data['kwargs'])

        obj.__init__(*positional_args, **keyword_args)
    return obj



from __future__ import annotations
from typing import Optional

from dictionizr import dictionize, undictionize

from .utils import eq


class Data():
    name: str
    data: Optional[Data] = None
    optional: Optional[str] = None
    def __init__(self, name: str, sub_name: Optional[str] = None):
        self.name = name
        if sub_name is not None:
            self.data = Data(sub_name)
    def __eq__(self, o: Data) -> bool:
        return eq(self, o)


def test__dictionize__property():
    data = Data('Joe')
    actual = dictionize(data)
    expected = { 'name': 'Joe' }
    assert expected == actual


def test__dictionize__no_functions():
    data = Data('Joe')
    actual = dictionize(data)
    expected = { 'name': 'Joe' }
    assert expected == actual


def test__dictionize__sub_objects():
    data = Data('Joe', 'John')
    actual = dictionize(data)
    expected = {
        'name': 'Joe',
        'data': { 'name': 'John'}
    }
    assert expected == actual


def test__undictionize__object():
    data = { 'name': 'Joe' }
    actual = undictionize(data, Data)
    expected = Data('Joe')
    assert expected == actual


def test__undictionize__sub_object():
    data = {
        'name': 'Joe',
        'data': { 'name': 'John'}
    }
    actual = undictionize(data, Data)
    expected = Data('Joe', 'John')
    assert expected == actual



def eq(a, b) -> bool:
    if a is None and b is None:
        return True
    elif a is None or b is None:
        return False

    try:
        a_vars = vars(a) 
        b_vars = vars(b)
        if len(a_vars) != len(b_vars):
            return False

        for a_key, a_value in a_vars.items():
            b_value = b_vars[a_key]
            if not eq(a_value, b_value):
                return False
        return True
    except TypeError:
        if isinstance(a, list) and isinstance(b, list):
            if len(a) != len(b):
                return False
            a = sorted(a)
            b = sorted(b)
            for index, a_value in enumerate(a):
                b_value = b[index]
                if not eq(a_value, b_value):
                    return False
            return True

        return a == b